# automated tester program for The Way of the Program
# written by Jim Marshall
# http://science.slc.edu/jmarshall
# last modified September 2023

# production version
__version__ = '1.13p' # same as 1.13, but with comments and extraneous code deleted

STRICT_NAMES = True

import io, sys, string, inspect, os

class MyInputStream(object):
   def __init__(self, s, output_stream, output_with_inputs):
      self._internal_sio = io.StringIO(s)
      self._output_stream = output_stream
      self._output_with_inputs = output_with_inputs
   def readline(self):
      result = self._internal_sio.readline()
      if len(result) == 0 or result[-1] != '\n':
         result += '\n'
      chars = self._output_stream.getvalue()
      if len(chars) > 0 and not chars[-1].isspace():
         self._output_stream.write(' ')
      self._output_with_inputs.write(result)
      return result

class MyOutputStream(object):
   def __init__(self, output_with_inputs):
      self._internal_sio = io.StringIO()
      self._output_with_inputs = output_with_inputs
   def write(self, s):
      self._output_with_inputs.write(s)
      return self._internal_sio.write(s)
   def getvalue(self):
      return self._internal_sio.getvalue()

class CapturedOutput(object):
   def __init__(self, console_input):
      self.console_input = console_input
   def __enter__(self):
      self._saved_stdin = sys.stdin
      self._saved_stdout = sys.stdout
      self._output_with_inputs = io.StringIO()
      sys.stdout = MyOutputStream(self._output_with_inputs)
      sys.stdin = MyInputStream(self.console_input, sys.stdout, self._output_with_inputs)
      return self
   def __exit__(self, *args):
      self.text_without_inputs = sys.stdout.getvalue()
      self.text_with_inputs = self._output_with_inputs.getvalue()
      sys.stdin = self._saved_stdin
      sys.stdout = self._saved_stdout

def plural(x):
   return "" if x == 1 else "s"

def number_of_arguments(function):
   return len(inspect.getfullargspec(function)[0])

def find_similar_name(x, names):
   for name in names:
      if x.lower() == name.lower():
         return name
   return None

def copy_args(args):
   import tkinter, copy
   return tuple(x if isinstance(x, tkinter.Canvas) else copy.deepcopy(x) for x in args)

def test_target_list(all, test_cases_table, start, default_mode, clear):
   if start is None:
      target_list = all
   elif type(start) is str:
      names = [x if type(x) is str else x['name'] for x in all]
      if start in names:
         target_list = all[names.index(start):]
      elif start == '?':
         print("Target list:\n%s" % ', '.join(dict.fromkeys(names)))
         return
      else:
         similar_name = find_similar_name(start, names)
         if STRICT_NAMES or similar_name is None:
            print("Sorry, %s not found in target list" % start)
            return
         else:
            start = similar_name
            target_list = all[names.index(start):]
   elif type(start) is type(lambda x: x):
      test_target_list(all, test_cases_table, start.__name__, default_mode, clear)
      return
   else:
      print("Sorry, expected a string or function object")
      return
   skipNext = False
   i = 0
   while i < len(target_list):
      if not skipNext:
         if clear and os.name == 'posix':
            os.system('clear')
         if type(target_list[i]) is str:
            name = target_list[i]
            test_single_function(name, 'all', default_mode, None, None, test_cases_table)
         elif type(target_list[i]) is dict:
            d = target_list[i]
            name, cases, mode = d['name'], d['cases'], d['mode']
            if mode is None:
               mode = default_mode
            test_single_function(name, cases, mode, None, None, test_cases_table)
      if i < len(target_list)-1:
         next = target_list[i+1]
         next_name = next if type(next) is str else next['name']
         if type(next) is str or next['cases'] == 'all':
            cases = ''
         elif type(next['cases']) is int:
            cases = ' case %s' % next['cases']
         else:
            cases = ' cases %s' % str(next['cases']).strip('[()]')
         action = input("\nPress RETURN to test %s%s (q: quit, s: skip, r: rerun, b: go back)..."
                        % (next_name, cases))
         if action in ('q', 'Q', 'quit'):
            return
         skipNext = action in ('s', 'S', 'skip')
         if action in ('b', 'B', 'back'):
            i = i - 2 if i > 0 else i - 1
         elif action in ('r', 'R', 'rerun'):
            i = i - 1
      i = i + 1

def opts(function_name, cases='all', mode=None):
   return {'name': function_name, 'cases': cases, 'mode': mode}

def process_args(x, cases, mode, alias, test_cases_table):
   if test_cases_table is None:
      print("Sorry, no test cases specified")
      return
   if not (cases is None or cases == 'all' or type(cases) in (int, list, tuple)):
      print("Sorry, cases must be 'all' or a valid list of case numbers")
      return
   if not (mode is None or mode in ('brief', 'verbose')):
      print("Sorry, mode must be 'brief' or 'verbose'")
      return
   if cases is None and mode is None:
      cases = 'all'
      mode = 'brief'
   elif cases is None and mode is not None:
      cases = 'all'
   elif cases is not None and mode is None:
      mode = 'verbose'
   if type(x) is type(lambda x: x):
      function_to_test = x
      name = function_to_test.__name__
      if type(alias) is str:
         name = alias
         if name not in test_cases_table:
            print("Sorry, no test cases available for %s" % name)
            return
   elif type(x) is str:
      main = sys.modules["__main__"]
      if x in dir(main):
         name = x
         function_to_test = main.__getattribute__(name)
      else:
         name = find_similar_name(x, dir(main))
         if name is None:
            print("Sorry, there is no function named '%s'" % x)
            return
         function_to_test = main.__getattribute__(name)
   else:
      print("Sorry, I don't know how to test that")
      return
   if name not in test_cases_table:
      similar_name = find_similar_name(name, test_cases_table.keys())
      if similar_name is None:
         print("Sorry, no test cases available for %s" % name)
         return
      if STRICT_NAMES:
         print("Sorry, your function %s must be named '%s'" % (name, similar_name))
         return
      else:
         main = sys.modules["__main__"]
         function_to_test = main.__getattribute__(name)
         name = similar_name
   all_cases = test_cases_table[name]
   if cases == 'all':
      casenums = sorted(all_cases.keys())
   elif type(cases) is int and cases in all_cases:
      casenums = [cases]
   elif type(cases) in (tuple, list) and all([x in all_cases for x in cases]):
      casenums = cases
   else:
      valid_casenums = "1" if len(all_cases) == 1 else "in the range 1-%d" % len(all_cases)
      print("Sorry, cases must be 'all' or %s" % valid_casenums)
      return
   return function_to_test, name, all_cases, casenums, mode

function_args = None

def test_single_function(x, cases, mode, alias, altinput, test_cases_table):
   global function_args
   results = process_args(x, cases, mode, alias, test_cases_table)
   if results is None: return
   function_to_test, name, all_cases, casenums, mode = results
   num_passed = 0
   num_failed = 0
   width = 75
   for casenum in casenums:
      case = all_cases[casenum]
      label = ' %s: case %d ' % (name, casenum)
      left_hyphens = '-' * ((width - len(label))//2)
      right_hyphens = '-' * (width - len(label) - len(left_hyphens))
      print('%s%s%s' % (left_hyphens, label, right_hyphens))
      if altinput is None:
         console_input = case[0]
      elif type(altinput) is str:
         console_input = altinput
      else:
         print("Sorry, input data must be a string")
         return
      if type(console_input) is not str:
         print("*** ERROR: invalid console input specified in test cases (must be a string)")
         num_failed += 1
         continue
      function_args = case[1]
      if type(function_args) == type(lambda x: x):
         function_args = function_args()
      if type(function_args) is not tuple:
         print("*** ERROR: invalid function args specified in test cases (must be a tuple)")
         num_failed += 1
         continue
      num_args = number_of_arguments(function_to_test)
      if num_args != len(function_args):
         print("*** ERROR: %s should take %d input parameter%s, but takes %d" %
               (name, len(function_args), plural(len(function_args)), num_args))
         num_failed += 1
         continue
      crashed = False
      with CapturedOutput(console_input) as output:
         try:
            function_args_copy = copy_args(function_args)
            return_value = function_to_test(*function_args_copy)
         except Exception as e:
            return_value = e
            crashed = True
      if crashed:
         show_output(name, function_args, output)
         print("*** ERROR: crashed with %s exception:" % type(return_value).__name__)
         print(return_value)
         num_failed += 1
         continue
      if len(case) == 2:
         if mode == 'verbose':
            show_output(name, function_args, output)
            if return_value is not None:
               print("returned %s\n" % repr(return_value))
         num_passed += 1
         continue
      expected_return_value = case[2]
      if type(expected_return_value) == type(lambda x: x):
         try:
            verifier_result = expected_return_value(return_value)
         except Exception as e:
            show_output(name, function_args, output)
            print("*** ERROR: returned the incorrect value %s" % truncate(repr(return_value)))
            num_failed += 1
            continue
         if verifier_result == True:
            pass
         elif verifier_result == False:
            show_output(name, function_args, output)
            print("*** ERROR: returned the incorrect value %s" % repr(return_value))
            num_failed += 1
            continue
         elif type(verifier_result) is str:
            show_output(name, function_args, output)
            print("*** ERROR: %s" % verifier_result)
            num_failed += 1
            continue
         else:
            show_output(name, function_args, output)
            print("*** ERROR: verifier function returned bogus value: %s" % verifier_result)
            num_failed += 1
            continue
      elif return_value != expected_return_value:
         show_output(name, function_args, output)
         print("*** ERROR: should return %s but returned %s" %
               (repr(expected_return_value), repr(return_value)))
         num_failed += 1
         continue
      else:
         pass
      expected_outputs = case[3:]
      if len(expected_outputs) == 0:
         if mode == 'verbose':
            show_output(name, function_args, output)
            print("returned %s\n" % repr(return_value))
         print("...passed")
         num_passed += 1
         continue
      if type(expected_outputs[0]) == type(lambda x: x):
         output_verifier_function = expected_outputs[0]
         result = output_verifier_function(output.text_without_inputs)
         if result == True:
            if mode == 'verbose':
               show_output(name, function_args, output)
            print("...passed")
            num_passed += 1
         else:
            show_output(name, function_args, output)
            print("*** ERROR: %s" % result)
            num_failed += 1
         continue
      unmatched = check_output(output.text_without_inputs, expected_outputs)
      if unmatched is None:
         if mode == 'verbose':
            show_output(name, function_args, output)
         print("...passed")
         num_passed += 1
      else:
         show_output(name, function_args, output)
         if type(unmatched) == Lit and '\n' in unmatched.string:
            print("*** ERROR: should be:\n%s" % unmatched)
         else:
            print("*** ERROR: should be \"%s\"" % unmatched)
         num_failed += 1
   print('=' * width)
   if len(casenums) == 1:
      pass
   elif num_passed == len(casenums):
      print("%s passed all tests" % name)
   else:
      print("%s passed %d test%s, %d problem%s detected" %
            (name, num_passed, plural(num_passed), num_failed, plural(num_failed)))
   
def show_function_call(name, function_args):
   args = tuple(repr(x) for x in function_args)
   args = (truncate(x) for x in args)
   args = ", ".join(args)
   print(">>> %s(%s)" % (name, args))

def truncate(x):
   if type(x) is not str:
      x = str(x)
   arglen_limit = 20
   if len(x) > arglen_limit:
      x = x[:arglen_limit] + "..."
      if x[0] == '[':
         x += ']'
      elif x[0] == '(':
         x += ')'
      elif x[0] == '{':
         x += '}'
      elif x[0] == '"':
         x += '"'
      elif x[0] == "'":
         x += "'"
   return x

def show_output(name, function_args, output):
   show_function_call(name, function_args)
   if len(output.text_with_inputs) > 0:
      print(output.text_with_inputs)

def check_output(output, expected_outputs):
   for x in expected_outputs:
      if type(x) in [tuple, list]:
         alternatives = x
      elif type(x) in [str, Lit]:
         alternatives = (x,)
      else:
         raise Exception("Improper test case value: %s" % x)
      output = any_match(alternatives, output)
      if output is None:
         return alternatives[0]
   return None

def any_match(alternatives, output):
   for a in alternatives:
      if type(a) is Lit:
         if a.string in output:
            start = output.index(a.string)
            end = start + len(a.string)
            remaining_output = output[end:]
            return remaining_output
      elif type(a) is str:
         normed_a = normalize(a).lower()
         normed_output = normalize(output).lower()
         if normed_a in normed_output:
            start = normed_output.index(normed_a)
            end = start + len(normed_a)
            token = normed_a.split()[0]
            original_start = find_in_original_text(token, output.lower())
            original_end = original_start + len(token)
            remaining_output = output[original_end:]
            return remaining_output
   return None

letters_or_digits = string.ascii_lowercase + string.ascii_uppercase + string.digits

def find_in_original_text(token, s):
   index = s.index(token)
   start = s.find(token)
   while start != -1:
      end = start + len(token)
      if ((start == 0 or s[start-1] not in letters_or_digits) and
          (end >= len(s) or s[end] not in letters_or_digits)):
         return start
      else:
         start = s.find(token, end)
   return index

def normalize(s):
   tokens = []
   for x in s.split():
      tokens.extend(detach_punctuation(x))
   return ' ' + ' '.join(tokens) + ' '

def detach_punctuation(s):
   if is_number(s):
      if s[-1] == '.':
         s = s[:-1]
      return [s]
   root = s.strip(string.punctuation)
   if root == '':
      return [s]
   start = s.index(root)
   end = start + len(root)
   result = []
   if start > 0:
      result.append(s[:start])
   result.append(root)
   if end < len(s):
      result.append(s[end:])
   return result

def is_number(s):
   try:
      num = float(s)
      return True
   except ValueError:
      return False

class Lit(object):
   def __init__(self, s):
      self.string = s
   def __str__(self):
      return self.string
   def __repr__(self):
      return 'Lit(%s)' % repr(self.string)
